/*****************************************************************************
 *
 * FILE:	netfd.c
 * DESCRIPTION:	Easy Network I/O library file
 * DATE:	Sat, Jun 10 2006
 * UPDATED:	Mon, Sep 17 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: netfd.c,v 1.11 2007/09/17 13:03:32 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif	/* HAVE_SYS_IOCTL_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <arpa/inet.h>
#include <sys/un.h>
#include <net/if_dl.h>
#if	HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif	/* HAVE_SYS_STAT_H */
#include <sys/mman.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/bpf.h>
#include <net/ethernet.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/tcpip.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <sys/time.h>
#include <sys/event.h>
#include <signal.h>
#include <errno.h>
#include "netfd.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	EOL	'\0'
#define	LF	'\012'
#define	CR	'\015'
#define	SPC	'\040'
#define	CRLF	"\015\012"

#if	EAGAIN != EWOULDBLOCK
#define	NETFD_IO_NOT_READY	((errno == EAGAIN || errno == EWOULDBLOCK))
#else
#define	NETFD_IO_NOT_READY	(errno == EAGAIN)
#endif

#define	xfree(x)	do { if ((x)!=NULL) { free((x)); (x)=NULL; } } while (0)

/* default value */
#define	DEF_MAXFDS		NETFD_DEFAULT_MODEL
#define	DEF_BACKLOG		(8)
#define	DEF_BLOCKMODE		(1)	/* 0: block, 1: nonblock */
#define	DEF_CONNECT_TIMEOUT	(5000)	/* msec */

/* netfd virtual processor */
typedef struct _netfd_vp_t {
  int	maxfds;
  int	curfds;
  NetFD	freeobj;
  struct {
    int			fd;	/* fro kqueu(2) */
    struct kevent *	events;
    int			nevents;
  } kq;
} netfd_vp_t;


#ifndef	max
#define	max(a,b)	((a)>(b)?(a):(b))
#endif	/* max */

/******************************************************************************
 *
 *	Lobal functions declaration
 *
 *****************************************************************************/
static NetFD	netfd_new(int osfd, int nonblock);
static void	netfd_free(NetFD nfd);
static ssize_t	one_read(NetFD nfd, char * ptr, time_t timeout);
static int	readbuf(NetFD nfd, time_t timeout);
static void	sigalrm(int signo);
static int	netfd_client(const char * hostname, const char * service,
			     int socktype, int protocol, int flags,
			     const time_t timeout,
			     struct sockaddr ** saddr, socklen_t * addrlen);
static int	netfd_server(const char * hostname, const char * service,
			     int socktype, int protocol, int flags,
			     struct sockaddr ** saddr, socklen_t * addrlen);
static int	my_close(int fd);
static char *	my_mktemp(char * path);
static char *	etherprint(const uint8_t ea[6], char s[NETFD_ETHER_ADDR_SIZE]);
static int	get_ifconf(int sd, struct ifconf * ifc, char ** buf);

/******************************************************************************
 *
 *	Lobal variable definition
 *
 *****************************************************************************/
static netfd_vp_t *		nvp = NULL;
static volatile sig_atomic_t	timedout = 0;	/* flag for connect() */ 
static sig_atomic_t		dispatched = 0;	/* flag for netfd_dispatch() */ 

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
int
netfd_init(maxfds)
	int	maxfds;
{
  if (nvp == NULL) {
    /* create netfd_vp_t object */
    nvp = (netfd_vp_t *)malloc(sizeof(netfd_vp_t));
    if (nvp != NULL) {
      /* regulation */
      if (maxfds <= 0) {
	maxfds = DEF_MAXFDS;
      }
      nvp->maxfds  = maxfds;
      nvp->curfds  = 0;
      nvp->freeobj = NULL;

      /* prepare kevent(2) */
      nvp->kq.nevents = 0;
      nvp->kq.events  = (struct kevent *)calloc((size_t)maxfds,
						sizeof(struct kevent));
      if (nvp->kq.events == NULL) {
	goto error1;
      }
      nvp->kq.fd = kqueue();
      if (nvp->kq.fd == -1) {
	goto error2;
      }
      return NETFD_INIT_OK;

error2:
      free(nvp->kq.events);
error1:
      free(nvp);
      nvp = NULL;
    }
  }
  else {
    return NETFD_INIT_ALREADY;	/* already initialized */
  }

  return NETFD_INIT_ERROR;
}

NetFD
netfd_open(path, flags, mode)
	const char *	path;
	int		flags;
	mode_t		mode;
{
  int	fd;
  NetFD	newfd;

  while ((fd = open(path, flags, mode)) < 0) {
    if (errno != EINTR) {
      return NULL;
    }
  }

  newfd = netfd_new(fd, DEF_BLOCKMODE);
  if (newfd == NULL) {
    my_close(fd);
  }

  return newfd;
}

NetFD
netfd_fdopen(osfd)
	int	osfd;
{
  return (osfd > 0) ? netfd_new(osfd, DEF_BLOCKMODE) : NULL;
}

static NetFD
netfd_new(osfd, nonblock)
	int	osfd;
	int	nonblock;
{
  NetFD	nfd;

  if (nvp->freeobj != NULL) {
    nfd		 = nvp->freeobj;
    nvp->freeobj = nvp->freeobj->next;
  }
  else if (nvp->curfds > nvp->maxfds) {
    errno = EMFILE;
    return NULL;
  }
  else {
    nfd = (NetFD)malloc(sizeof(netfd_t));
    if (nfd == NULL) {
      errno = ENOMEM;
      return NULL;
    }
    nvp->curfds++;
  }

  /* initialize */
  nfd->osfd	= osfd;
  nfd->rb.total = 0;
  nfd->rb.bufp	= nfd->rb.buf;
  memset(nfd->rb.buf, 0, sizeof(nfd->rb.buf));
  nfd->path	= NULL;
  nfd->addrlen	= 0;
  nfd->saddr	= NULL;
  nfd->flags	= 0;
  nfd->next	= NULL;
#if	DEBUG
  nfd->count	= 0;
#endif	/* DEBUG */

  if (nonblock) {
    int	flags;

    if ((flags = fcntl(osfd, F_GETFL, 0)) == -1 ||
	fcntl(osfd, F_SETFL, flags|O_NONBLOCK) == -1) {
      netfd_free(nfd);
      nfd = NULL;
    }
  }

  return nfd;
}

static void
netfd_free(nfd)
	NetFD	nfd;
{
  if (nfd != NULL) {
#if	DEBUG
    fprintf(stderr, "DEBUG[netfd_free] count=%u", nfd->count);
    if (nfd->flags == NETFD_FL_LISTENER) {
      fputs("[*]", stderr);
    }
    fputc('\n', stderr);
    nfd->count	     = 0;
#endif	/* DEBUG */
    if (nfd->path != NULL) {
      unlink(nfd->path);
      free(nfd->path);
    }
    if (nfd->saddr != NULL) {
      free(nfd->saddr);
    }
    nfd->next	  = nvp->freeobj;
    nvp->freeobj  = nfd;
  }
}

int
netfd_close(nfd)
	NetFD	nfd;
{
  int	status = -1;

  if (nfd == NULL) {
    errno  = EINVAL;
  }
  else {
    status = my_close(nfd->osfd);
    netfd_free(nfd);
  }

  return status;
}

off_t
netfd_seek(nfd, offset, whence)
	NetFD	nfd;
	off_t	offset;
	int	whence;
{
  struct stat	sbuf;

  if (fstat(nfd->osfd, &sbuf) == 0 && S_ISREG(sbuf.st_mode)) {
    off_t	pos;

    pos = lseek(nfd->osfd, offset, whence);
    if (pos != -1) {
      return pos;
    }
    errno = EINVAL;
  }

  return -1;
}

ssize_t
netfd_read(nfd, buf, nbytes, timeout)
	NetFD		nfd;
	void *		buf;
	size_t		nbytes;
	const time_t	timeout;	/* milliseconds */
{
  register ssize_t	n   = 0;
  register char *	ptr = (char *)buf;

  do {
    if (readbuf(nfd, timeout) < 0) { return -1; }
    if (nfd->rb.total == 0)	   { break; }
    *ptr++ = *nfd->rb.bufp++;
    nfd->rb.total--;
  } while ((size_t)n++ < nbytes && nfd->rb.total > 0);

  return n;
}

ssize_t
netfd_write(nfd, buf, nbytes, timeout)
	NetFD		nfd;
	const void *	buf;
	size_t		nbytes;
	const time_t	timeout;	/* milliseconds */
{
  size_t	rest = nbytes;
  ssize_t	n;

  while (rest > 0) {
    n = write(nfd->osfd, buf, rest);
    if (n == -1) {
      pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

      if (errno == EINTR)			{ continue; }
      if (!NETFD_IO_NOT_READY)			{ return -1; }
      if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
    }
    else {
      rest -= n;
      if (rest == 0) { break; }
      buf = (const void *)((const char *)buf + n);
    }
  }

  return (ssize_t)(nbytes - rest);
}

ssize_t
netfd_readline(nfd, buf, maxlen, timeout)
	NetFD		nfd;
	void *		buf;
	size_t		maxlen;
	const time_t	timeout;	/* milliseconds */
{
  register ssize_t	n;
  register char *	ptr = (char *)buf;

  for (n = 1; (size_t)n < maxlen; n++) {
    char	c;
    int		status;

    status = one_read(nfd, &c, timeout);
    if (status == 1) {
      *ptr++ = c;
      if (c == LF)  { break; }
      if (c == EOL) { return 0; }	/* EOF?, null data read */
    }
    else if (status == 0) {
      if (n == 1) { return 0; }		/* EOF, no data read */
      else	  { break; }		/* EOF, some data was read */
    }
    else {
      return -1;			/* error */
    }
  }
  *ptr = EOL;

  return n;
}

static ssize_t
one_read(nfd, ptr, timeout)
	NetFD		nfd;
	char *		ptr;
	const time_t	timeout;	/* milliseconds */
{
  if (readbuf(nfd, timeout) < 0) { return -1; }
  if (nfd->rb.total == 0)	 { return  0; }
  nfd->rb.total--;
  *ptr = *nfd->rb.bufp++;

  return 1;
}

static int
readbuf(nfd, timeout)
	NetFD		nfd;
	const time_t	timeout;	
{
  if (nfd->rb.total <= 0) {
    memset(nfd->rb.buf, 0, sizeof(nfd->rb.buf));
    while ((nfd->rb.total = read(nfd->osfd,
				 nfd->rb.buf, sizeof(nfd->rb.buf))) < 0) {
      pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };

      if (errno == EINTR)			{ continue; }
      if (!NETFD_IO_NOT_READY)			{ return -1; }
      if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
    }
    nfd->rb.bufp = nfd->rb.buf;
  }

  return nfd->rb.total;
}

/*****************************************************************************/

ssize_t
netfd_readv(nfd, iov, iovcnt, timeout)
	NetFD			nfd;
	const struct iovec *	iov;
	int			iovcnt;
	const time_t		timeout;
{
  ssize_t	n;

  while ((n = readv(nfd->osfd, iov, iovcnt)) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return -1; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
  }

  return n;
}

ssize_t
netfd_writev(nfd, iov, iovcnt, timeout)
	NetFD			nfd;
	const struct iovec *	iov;
	int			iovcnt;
	const time_t		timeout;
{
  int		i;
  ssize_t	nbytes;
  size_t	resid;
  ssize_t	n;

  for (nbytes = 0, i = 0; i < iovcnt; i++) {
    nbytes += iov[i].iov_len;
  }

  resid = nbytes;
  while (resid > 0) {
    n = writev(nfd->osfd, iov, iovcnt);
    if (n == -1) {
      pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

      if (errno == EINTR)			{ continue; }
      if (!NETFD_IO_NOT_READY)			{ return -1; }
      if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
    }
    else {
      resid -= n;
      break;
    }
  }

  if (resid > 0) {
    for (i = 0; i < iovcnt; i++) {
      if ((size_t)n >= iov[i].iov_len) {	/* written */
	n -= iov[i].iov_len;
      }
      else {
	ssize_t	l;

	l = netfd_write(nfd, iov[i].iov_base, iov[i].iov_len, timeout);
	if (l == -1) {
	  break;
	}
	n     -= l;
	resid -= l;
      }
    }
  }

  return (ssize_t)(nbytes - resid);
}

/*****************************************************************************/

ssize_t
netfd_recvfrom(nfd, buf, len, flags, from, fromlen, timeout)
	NetFD			nfd;
	void *			buf;
	size_t			len;
	int			flags;
	struct sockaddr *	from;
	socklen_t *		fromlen;
	const time_t		timeout;	/* milliseconds */
{
  ssize_t	n;

  while ((n = recvfrom(nfd->osfd, buf, len, flags, from, fromlen)) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return -1; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
  }

  return n;
}

ssize_t
netfd_sendto(nfd, msg, len, flags, to, tolen, timeout)
	NetFD			nfd;
	const void *		msg;
	size_t			len;
	int			flags;
	const struct sockaddr *	to;
	socklen_t		tolen;
	const time_t		timeout;	/* milliseconds */
{
  ssize_t	n;

  while ((n = sendto(nfd->osfd, msg, len, flags, to, tolen)) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return -1; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
  }

  return n;
}

NetFD
netfd_mkstemp(template)
	char *	template;
{
  int	fd;

  fd = mkstemp(template);
  return (fd == -1) ? NULL : netfd_new(fd, DEF_BLOCKMODE);
}


#if	HAVE_FREEBSD_SENDFILE
int
netfd_sendfile(nfd, nsd, init_offset, nbytes, hdtr, ret_sbytes, flags, timeout)
	NetFD			nfd;
	NetFD			nsd;
	off_t			init_offset;
	size_t			nbytes;
	struct sf_hdtr *	hdtr;
	off_t *			ret_sbytes;
	int			flags;
	const time_t		timeout;	/* milliseconds */
{
  off_t	offset;
  off_t	sbytes;

  offset = init_offset;
  while (sendfile(nfd->osfd, nsd->osfd, offset, nbytes, hdtr, &sbytes,
		  flags) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

    offset += sbytes;
    nbytes -= sbytes;

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return -1; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
  }
  if (ret_sbytes != NULL) {
    *ret_sbytes = offset + sbytes - init_offset;
  }

  return 0;
}
#else
#if	HAVE_LINUX_SENDFILE
/*
 * I have no linux environments. Is this routine ok?
 */
int
netfd_sendfile(nfd, nsd, init_offset, nbytes, hdtr, ret_sbytes, flags, timeout)
	NetFD		nfd;
	NetFD		nsd;
	off_t		init_offset;
	size_t		nbytes;
	void *		hdtr;
	off_t *		ret_sbytes;
	int		flags;
	const time_t	timeout;	/* milliseconds */
{
  off_t	offset;
  off_t	sbytes;

  offset = init_offset;
  while (sendfile(nfd->osfd, nsd->osfd, &sbytes, nbytes) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

    offset += sbytes;
    nbytes -= sbytes;

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return -1; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
  }
  if (ret_sbytes != NULL) {
    *ret_sbytes = offset + sbytes - init_offset;
  }

  return 0;
}
#else	/* No sendfile(2) */
int
netfd_sendfile(nfd, nsd, offset, nbytes, hdtr, sbytes, flags, timeout)
	NetFD		nfd;
	NetFD		nsd;
	off_t		offset;
	size_t		nbytes;
	void *		hdtr __unused;
	off_t *		sbytes;
	int		flags __unused;
	const time_t	timeout;	/* milliseconds */
{
  void *	buf;
  size_t	len;
  int		n;
  int		status = 0;

  if (nbytes == 0) {
    /* get whole file size */
    n = netfd_seek(nfd, (off_t)0, NETFD_SEEK_END);
    if (n == -1) {
      return -1;
    }
    len = (size_t)n;
  }
  else {
    len = nbytes;
  }
  buf = mmap(NULL, len, PROT_READ, MAP_PRIVATE, nfd->osfd, offset);
  if (buf == MAP_FAILED) {
    return -1;
  }
  n = netfd_write(nsd, buf, len, timeout);
  if (n == -1) {
    status = -1;
  }
  if (sbytes != NULL) {
    *sbytes = n;
  }
  munmap(buf, len);

  return status;
}
#endif	/* HAVE_LINUX_SENDFILE */
#endif	/* HAVE_FREEBSD_SENDFILE */

/*****************************************************************************/

NetFD
netfd_socket(domain, type, protocol)
	int	domain;
	int	type;
	int	protocol;
{
  int	sd;

  sd = socket(domain, type, protocol);
  if (sd != -1) {
    NetFD	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

NetFD
netfd_accept(nfd, addr, addrlen, timeout)
	NetFD			nfd;
	struct sockaddr *	addr;
	socklen_t *		addrlen;
	const time_t		timeout;	/* milliseconds */
{
  int		sd;
  NetFD		newfd;
#ifdef	SO_NOSIGPIPE
  const int	on = 1;
#endif	/* SO_NOSIGPIPE */

  while ((sd = accept(nfd->osfd, addr, addrlen)) < 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };

    if (errno == EINTR)				{ continue; }
    if (!NETFD_IO_NOT_READY)			{ return NULL; }
    if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return NULL; }
  }

#ifdef	SO_NOSIGPIPE
  /* Do not generate SIGPIPE */
  (void)setsockopt(sd, SOL_SOCKET, SO_NOSIGPIPE, &on, sizeof(on));
#endif	/* SO_NOSIGPIPE */

  newfd = netfd_fdopen(sd);
  if (newfd == NULL) {
    my_close(sd);
  }
  else {
    newfd->saddr = (struct sockaddr *)malloc(*addrlen);
    if (newfd->saddr != NULL) {
      memcpy(newfd->saddr, addr, *addrlen);
      newfd->addrlen = *addrlen;
    }
    else {
      netfd_close(newfd);
      newfd = NULL;
    }
  }

  return newfd;
}

/*****************************************************************************/

int
netfd_poll(pfds, pfdnum, timeout)
	pollfd_t	pfds[];
	size_t		pfdnum;
	const time_t	timeout;	/* milliseconds */
{
  int	n = -1;	/* return value */

  if (pfdnum > 0) {
    struct kevent *	events;

    events = (struct kevent *)calloc(pfdnum, sizeof(struct kevent));
    if (events != NULL) {
      int		nevents;
      struct timespec	to;
      struct timespec *	tsp;
      struct timeval	tv_st;
      register size_t	i;

      nevents = 0;
      for (i = 0; i < pfdnum; i++) {
	if (pfds[i].events & NETFD_POLLIN) {
	  EV_SET(&events[nevents++], pfds[i].nfd->osfd,
		 EVFILT_READ, EV_ADD | EV_ONESHOT, 0, 0, pfds[i].nfd);
	}
	if (pfds[i].events & NETFD_POLLOUT) {
	  EV_SET(&events[nevents++], pfds[i].nfd->osfd,
		 EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, pfds[i].nfd);
	}
	pfds[i].revents = 0;
      }

      if (timeout == 0) {
	to.tv_sec	= 0L;
	to.tv_nsec	= 0L;
	tsp		= &to;
      }
      else if (timeout < 0) {
	tsp = NULL;
      }
      else {
	to.tv_sec	= NETFD_MSEC2SEC(timeout);
	to.tv_nsec	= NETFD_MSEC2NSEC(timeout);
	tsp		= &to;
      }

      gettimeofday(&tv_st, (struct timezone *)NULL);
      while (((n = kevent(nvp->kq.fd,
    			  events, nevents, events, (int)pfdnum, tsp)) < 0) &&
    	    (errno == EINTR)) {
	/* re-calculate timeout */
	if (tsp != NULL) {
	  struct timeval	tv_ed;
	  struct timeval	tv_sub;

	  gettimeofday(&tv_ed, (struct timezone *)NULL);
	  timersub(&tv_ed, &tv_st, &tv_sub);
	  tv_st = tv_ed;
	  to.tv_sec  -= tv_sub.tv_sec;
	  to.tv_nsec -= (tv_sub.tv_usec * 1000);
	  if (to.tv_nsec < 0) {
	    to.tv_sec--;
	    to.tv_nsec += 1000000000;
	  }
	  if (to.tv_sec < 0) {
	    to.tv_sec = 0;
	  }
	}
	errno = 0;
      }
      switch (n) {
	case 0:	/* timeout */
	  errno = ETIMEDOUT;
	  break;
	case -1:	/* error */
	  if (errno == EBADF) {
	    for (i = 0; i < pfdnum; i++) {
	      pfds[i].revents = 0;
	      if (fcntl(pfds[i].nfd->osfd, F_GETFL, 0) == -1) {
		pfds[i].revents = NETFD_POLLNVAL;
	      }
	    }
	  }
	  break;
	default:
	  for (i = 0; i < (size_t)n; i++) {
	    if ((pfds[i].events & NETFD_POLLIN) &&
		events[i].ident  == pfds[i].nfd->osfd &&
		events[i].filter == EVFILT_READ) {
	      pfds[i].revents |= NETFD_POLLIN;
	    }
	    if ((pfds[i].events & NETFD_POLLOUT) &&
		events[i].ident  == pfds[i].nfd->osfd &&
		events[i].filter == EVFILT_WRITE) {
	      pfds[i].revents |= NETFD_POLLOUT;
	    }
	  }
	  break;
      }
      free(events);
    }
  }

  return n;
}

/*****************************************************************************/

NetFD
netfd_tcp_client(hostname, service, timeout)
	const char *	hostname;
	const char *	service;
	const time_t	timeout;	/* milliseconds */
{
  int			sd;
  struct sockaddr *	saddr;
  socklen_t		addrlen;

  sd = netfd_client(hostname, service, SOCK_STREAM, IPPROTO_TCP, 0, timeout,
		    &saddr, &addrlen);
  if (sd != -1) {
    NetFD	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      nfd->addrlen = addrlen;
      nfd->saddr   = saddr;
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

NetFD
netfd_tcp_server(hostname, service, backlog)
	const char *	hostname;
	const char *	service;
	const int	backlog;
{
  int			sd;
  struct sockaddr *	saddr;
  socklen_t		addrlen;

  sd = netfd_server(hostname, service, SOCK_STREAM, IPPROTO_TCP, AI_PASSIVE,
		    &saddr, &addrlen);
  if (sd != -1) {
    if (listen(sd, backlog > 0 ? backlog : DEF_BACKLOG) == 0) {
      NetFD	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	nfd->addrlen = addrlen;
	nfd->saddr   = saddr;
	nfd->flags  |= NETFD_FL_LISTENER;
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

/*****************************************************************************/

NetFD
netfd_udp_client(hostname, service, broadcast)
	const char *	hostname;
	const char *	service;
	bool		broadcast;
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  int			sd	= -1;
  struct sockaddr *	saddr	= NULL;
  socklen_t		addrlen = 0;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_family	= broadcast ? AF_INET : AF_UNSPEC;
  hints.ai_socktype	= SOCK_DGRAM;
  hints.ai_protocol	= IPPROTO_UDP;
  hints.ai_flags	= 0;

  if (getaddrinfo(hostname, service, &hints, &res) == 0) {
    struct addrinfo *	res_save;

    res_save = res;
    do {
      /* create a socket */
      sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
      if (sd != -1) {
	if (broadcast) {
	  const int	on = 1;
	  if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)) == -1) {
	    goto error;
	  }
	}
	addrlen	= res->ai_addrlen;
	saddr	= malloc(res->ai_addrlen);
	if (saddr != NULL) {
	  memcpy(saddr, res->ai_addr, res->ai_addrlen);
	  break;
	}
      }
  error:
      my_close(sd);	/* ignore this one */
      sd = -1;
    } while ((res = res->ai_next) != NULL && errno == 0);
    freeaddrinfo(res_save);
  }

  if (sd != -1) {
    NetFD	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      nfd->addrlen = addrlen;
      nfd->saddr   = saddr;
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

NetFD
netfd_udp_connect(hostname, service)
	const char *	hostname;
	const char *	service;
{
  int			sd;
  struct sockaddr *	saddr;
  socklen_t		addrlen;

  sd = netfd_client(hostname, service, SOCK_DGRAM, IPPROTO_UDP, 0,
		    DEF_CONNECT_TIMEOUT, &saddr, &addrlen);
  if (sd != -1) {
    NetFD	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      nfd->addrlen = addrlen;
      nfd->saddr   = saddr;
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

NetFD
netfd_udp_server(hostname, service)
	const char *	hostname;
	const char *	service;
{
  int			sd;
  struct sockaddr *	saddr;
  socklen_t		addrlen;

  sd = netfd_server(hostname, service, SOCK_DGRAM, IPPROTO_UDP, AI_PASSIVE,
		    &saddr, &addrlen);
  if (sd != -1) {
    NetFD	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      nfd->addrlen = addrlen;
      nfd->saddr   = saddr;
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

NetFD
netfd_udp_open2(hostname, connport, waitport, family, broadcast, to, tolen)
	const char *		hostname;	/* remote host */
	const char *		connport;	/* remote port */
	const char *		waitport;	/* local port */
	int			family;		/* see <sys/socket.h> */
	bool			broadcast;
	struct sockaddr **	to;		/* returned */
	socklen_t *		tolen;		/* returned */
{
  struct addrinfo	hints;
  struct addrinfo *	server;
  struct addrinfo *	client;
  int			sd = -1;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_family	= broadcast ? AF_INET : family;
  hints.ai_socktype	= SOCK_DGRAM;
  hints.ai_protocol	= IPPROTO_UDP;
  hints.ai_flags	= 0;

  /* server */
  if (getaddrinfo(hostname, connport, &hints, &server) == 0) {
    struct addrinfo *	res_save;

    res_save = server;
    do {
      /* create a socket */
      sd = socket(server->ai_family, server->ai_socktype, server->ai_protocol);
      if (sd != -1) {
	const int	on = 1;

	(void)setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	if (broadcast) {
	  if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)) == -1) {
	    goto error;
	  }
	}
	*tolen	= server->ai_addrlen;
	*to	= malloc(server->ai_addrlen);
	if (*to != NULL) {
	  memcpy(*to, server->ai_addr, server->ai_addrlen);
	  break;
	}
      }
  error:
      my_close(sd);	/* ignore this one */
      sd = -1;
    } while ((server = server->ai_next) != NULL && errno == 0);
    freeaddrinfo(res_save);
  }

  /* client */
  if (sd != -1) {
    hints.ai_flags = AI_PASSIVE;
    if (getaddrinfo(NULL, waitport, &hints, &client) == 0) {
      struct addrinfo *	res_save;
      int		rc = -1;

      res_save = client;
      do {
	if ((rc = bind(sd, client->ai_addr, client->ai_addrlen)) == 0) {
	  break;
	}
      } while ((client = client->ai_next) != NULL && errno == 0);
      freeaddrinfo(res_save);
      if (rc == -1) {
	my_close(sd);	/* ignore this one */
	sd = -1;
      }
    }
  }

  if (sd != -1) {
    NetFD	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

/*****************************************************************************/

NetFD
netfd_ust_client(path)
	const char *	path;
{
  int	sd;

  sd = socket(AF_LOCAL, SOCK_STREAM, 0);
  if (sd != -1) {
    struct sockaddr_un	caddr;

    memset((char *)&caddr, 0, sizeof(caddr));
    caddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
    strlcpy(caddr.sun_path, path, sizeof(caddr.sun_path));
#else
    strncpy(caddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
    if (connect(sd, (struct sockaddr *)&caddr, sizeof(caddr)) == 0) {
      NetFD	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

NetFD
netfd_ust_server(path, backlog)
	const char *	path;
	const int	backlog;
{
  int	sd;

  sd = socket(AF_LOCAL, SOCK_STREAM, 0);
  if (sd != -1) {
    struct sockaddr_un	saddr;

    unlink(path);
    memset((char *)&saddr, 0, sizeof(saddr));
    saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
    strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
    strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
    if (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == 0 &&
	listen(sd, backlog > 0 ? backlog : DEF_BACKLOG) == 0) {
      NetFD	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

/*****************************************************************************/

NetFD
netfd_udg_client(path)
	const char *	path;
{
  char	tmpath[] = "/tmp/.s.udg_client.XXXXXX";

  if (my_mktemp(tmpath) != NULL) {
    int	sd;

    sd = socket(AF_LOCAL, SOCK_DGRAM, 0);
    if (sd != -1) {
      struct sockaddr_un	caddr;

      memset((char *)&caddr, 0, sizeof(caddr));
      caddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
      strlcpy(caddr.sun_path, tmpath, sizeof(caddr.sun_path));
#else
      strncpy(caddr.sun_path, tmpath, strlen(tmpath));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
      if (bind(sd, (struct sockaddr *)&caddr, sizeof(caddr)) == 0) {
	struct sockaddr_un	saddr;

	memset((char *)&saddr, 0, sizeof(saddr));
	saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
	strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
	strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
	if (connect(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == 0) {
	  NetFD	nfd;

	  nfd = netfd_fdopen(sd);
	  if (nfd != NULL) {
	    nfd->path = strdup(tmpath);
	    return nfd;
	  }
	}
      }
      my_close(sd);
    }
  }

  return NULL;
}

NetFD
netfd_udg_server(path)
	const char *	path;
{
  int	sd;

  sd = socket(AF_LOCAL, SOCK_DGRAM, 0);
  if (sd != -1) {
    struct sockaddr_un	saddr;

    unlink(path);
    memset((char *)&saddr, 0, sizeof(saddr));
    saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
    strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
    strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
    if (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == 0) {
      NetFD	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

/* cf. /usr/src/lib/libc/stdio/mktemp.c */
static char *
my_mktemp(path)
	char *	path;
{
  static const unsigned char	padchar[] =
	"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
  char *	retval = path;
  char *	trv;
  uint32_t	rnd;

  for (trv = path; *trv != '\0'; trv++) { continue; }
  if (--trv < path) {
    errno = EINVAL;
    return NULL;
  }

  while (trv >= path && *trv == 'X') {
    rnd = arc4random() % (sizeof(padchar) - 1);
    *trv-- = padchar[rnd];
  }

  return retval;
}

/*****************************************************************************/

static void
sigalrm(signo)
	int	signo __unused;
{
  timedout = 1;	/* set flag */
#if	DEBUG
  fprintf(stderr, "DEBUG[sigalrm]\n");
#endif	/* DEBUG */
}

static int
netfd_client(hostname, service, socktype, protocol, flags, timeout,
	     saddr, addrlen)
	const char *		hostname;
	const char *		service;
	int			socktype;
	int			protocol;
	int			flags;
	const time_t		timeout;	/* milliseconds */
	struct sockaddr **	saddr;		/* returned */
	socklen_t *		addrlen;	/* returned */
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  int			sd = -1;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= socktype;
  hints.ai_protocol	= protocol;
  hints.ai_flags	= flags;

  if (getaddrinfo(hostname, service, &hints, &res) == 0) {
    struct addrinfo *	res_save;
    struct itimerval	value;	/* new */
    struct itimerval	ovalue;	/* old */
    struct sigaction	act;	/* new */
    struct sigaction	oact;	/* old */

    sigemptyset(&act.sa_mask);
    act.sa_flags	= 0;
    act.sa_handler	= (void *)sigalrm;
    sigaction(SIGALRM, &act, &oact);
    value.it_interval.tv_sec = value.it_interval.tv_usec = 0;
    value.it_value.tv_sec  = NETFD_MSEC2SEC(timeout);
    value.it_value.tv_usec = NETFD_MSEC2USEC(timeout);
    setitimer(ITIMER_REAL, &value, &ovalue);
    timedout = 0;	/* clear flag */

    res_save = res;
    do {
      /* create a socket */
      sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
      if (sd == -1) {
	continue;	/*ignore this one */
      }
      /* initiate a connection on a socket */
      errno = 0;
      if (connect(sd, res->ai_addr, res->ai_addrlen) == 0) {
	*addrlen = res->ai_addrlen;
	*saddr	  = (struct sockaddr *)malloc(res->ai_addrlen);
	if (*saddr != NULL) {
	  memcpy(*saddr, res->ai_addr, res->ai_addrlen);
	}
	break;
      }
      if (errno == EINTR) {
	if (timedout) {
	  errno = ETIMEDOUT;
#if	DEBUG
	  fprintf(stderr, "DEBUG[netfd_client] timedout\n");
#endif	/* DEBUG */
	}
      }
      my_close(sd);	/* ignore this one */
      sd = -1;
    } while ((res = res->ai_next) != NULL && errno == 0);
    freeaddrinfo(res_save);
    sigaction(SIGALRM, &oact, NULL);
    setitimer(ITIMER_REAL, &ovalue, NULL);
  }

  return sd;
}

static int
netfd_server(hostname, service, socktype, protocol, flags, saddr, addrlen)
	const char *		hostname;
	const char *		service;
	int			socktype;
	int			protocol;
	int			flags;
	struct sockaddr **	saddr;		/* returned */
	socklen_t *		addrlen;	/* returned */
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  int			sd = -1;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= socktype;
  hints.ai_protocol	= protocol;
  hints.ai_flags	= flags;

  if (getaddrinfo(hostname, service, &hints, &res) == 0) {
    struct addrinfo *	res_save;

    res_save = res;
    do {
      static const int	on = 1;

      /* create a socket */
      sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
      if (sd == -1) {
	continue;	/*error, try next one */
      }
      /* allow reuse of local address */
      (void)setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

      /* bind a socket */
      if (bind(sd, res->ai_addr, res->ai_addrlen) == 0) {
	*addrlen = res->ai_addrlen;
	*saddr	  = (struct sockaddr *)malloc(res->ai_addrlen);
	if (*saddr != NULL) {
	  memcpy(*saddr, res->ai_addr, res->ai_addrlen);
	}
	break;	/* success */
      }
      my_close(sd);	/* bind error, close and try next one */
      sd = -1;
    } while ((res = res->ai_next) != NULL);
    freeaddrinfo(res_save);
  }

  return sd;
}

/*****************************************************************************/

static int
my_close(fd)
	int	fd;
{
  int	status;
  int	saved_errno = errno;
  while (((status = close(fd)) < 0) && (errno == EINTR)) { continue; }
  errno = saved_errno;
  return status;
}

/*****************************************************************************/

int
netfd_host_serv(hostname, service, family, socktype, protocol, res)
	const char *		hostname;
	const char *		service;
	int			family;
	int			socktype;
	int			protocol;
	struct addrinfo **	res;
{
  struct addrinfo	hints;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= AI_CANONNAME;	/* always return canonical name */
  hints.ai_family	= family;
  hints.ai_socktype	= socktype;
  hints.ai_protocol	= protocol;

  return getaddrinfo(hostname, service, &hints, res);
}

int
netfd_getnameinfo(nfd, host, hostlen, serv, servlen)
	NetFD	nfd;
	char *		host;
	size_t		hostlen;
	char *		serv;
	size_t		servlen;
{
  struct sockaddr	name;
  socklen_t		namelen = sizeof(name);

  if (getpeername(nfd->osfd, &name, &namelen) == -1) {
    return -1;
  }
  if (getnameinfo(&name, namelen, host, hostlen, serv, servlen,
		  NI_NUMERICHOST | NI_NUMERICSERV) == -1) {
    return -1;
  }
  return 0;
}

#if	0
char *
netfd_sock_ntop(sa, has_port)
	const struct sockaddr *	sa;
	bool			has_port;	/* 1: need */
{
  char	port[7];
  char	host[128];	/* Unix domain is largest */

  memset(port, 0, sizeof(port));
  memset(host, 0, sizeof(host));

  switch (sa->sa_family) {
    case 255: {
	unsigned long	mask;
	unsigned int	index0 = 1 << 31;
	unsigned short	new_mask = 0;

	mask = ntohl(((const struct sockaddr_in *)sa)->sin_addr.s_addr);
	while (mask & index0) {
	  new_mask++;
	  index0 >>= 1;
	}
#if	HAVE_SNPRINTF
	if (snprintf(host, sizeof(host), "/%hu", new_mask) == -1) {
	  return NULL;
	}
#else
	if (sprintf(host, "/%hu", new_mask) == -1) {
	  return NULL;
	}
#endif	/* HAVE_SNPRINTF */
      }
      break;

    case AF_UNSPEC:
    case AF_INET: {
	const struct sockaddr_in *	sin = (const struct sockaddr_in *)sa;

	if (inet_ntop(AF_INET, &sin->sin_addr, host, sizeof(host)) == NULL) {
	  return NULL;
	}
	if (strcmp(host, "0.0.0.0") == 0 && has_port == false) {
#if	HAVE_STRLCPY
	  strlcpy(host, "default", sizeof(host));
#else
	  strncpy(host, "default", 7);	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
	}
	if (ntohs(sin->sin_port) != 0 && has_port == true) {
#if	HAVE_SNPRINTF
	  if (snprintf(port, sizeof(port), ".%d", ntohs(sin->sin_port)) == -1) {
	    return NULL;
	  }
#else
	  int	sport = ntohs(sin->sin_port);
	  if (sport > 99999) {	/* over the number with five figures */
	    return NULL;
	  }
	  if (sprintf(port, ".%d", sport) == -1) {
	    return NULL;
	  }
#endif	/* HAVE_SNPRINTF */
#if	HAVE_STRLCAT
	  strlcat(host, port, sizeof(host));
#else
	  strncat(host, port, strlen(port));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCAT */
	}
      }
      break;

    case AF_INET6: {
	const struct sockaddr_in6 *	sin6 = (const struct sockaddr_in6 *)sa;

	if (inet_ntop(AF_INET6, &sin6->sin6_addr, host, sizeof(host)) == NULL) {
	  return NULL;
	}
	if (ntohs(sin6->sin6_port) != 0 && has_port == true) {
#if	HAVE_SNPRINTF
	  if (snprintf(port, sizeof(port), ".%d", ntohs(sin6->sin6_port)) == -1) {
	    return NULL;
	  }
#else
	  int	sport = ntohs(sin6->sin6_port);
	  if (sport > 99999) {	/* over the number with five figures */
	    return NULL;
	  }
	  if (sprintf(port, ".%d", ntohs(sin6->sin6_port)) == -1) {
	    return NULL;
	  }
#endif	/* HAVE_SNPRINTF */
#if	HAVE_STRLCAT
	  strlcat(host, port, sizeof(host));
#else
	  strncat(host, port, strlen(port));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCAT */
	}
      }
      break;

    case AF_UNIX: {
	const struct sockaddr_un *	unp = (const struct sockaddr_un *)sa;

	if (unp->sun_path[0] == EOL) {
#if	HAVE_STRLCPY
	  strlcpy(host, "(no pathname found)", sizeof(host));
#else
	  strncpy(host, "(no pathname found)", 19);	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
	}
	else {
#if	HAVE_SNPRINTF
	  if (snprintf(host, sizeof(host), "%s", unp->sun_path) == -1) {
	    return NULL;
	  }
#else
	  if (sprintf(host, "%.*s", (int)(sizeof(host) - 1), unp->sun_path) == -1) {
	    return NULL;
	  }
#endif	/* HAVE_SNPRINTF */
	}
      }
      break;

#if	AF_LINK
    case AF_LINK: {
	const struct sockaddr_dl *	sdl = (const struct sockaddr_dl *)sa;

	if (sdl->sdl_nlen > 0) {
	  memset(host, 0, sizeof(host));
	  memcpy(host, sdl->sdl_data, sdl->sdl_nlen);
	}
	else {
#if	HAVE_SNPRINTF
	  snprintf(host, sizeof(host), "link#%d", sdl->sdl_index);
#else
	  sprintf(host, "link#%d", sdl->sdl_index);
#endif	/* HAVE_SNPRINTF */
	}
      }
      break;
#endif	/* AF_LINK */

    default:
      return NULL;
  }

  return strdup(host);
}
#else
char *
netfd_sock_ntop(sa0, with_port, dst, size)
	const struct sockaddr *	sa0;
	bool			with_port;
	char *			dst;
	size_t			size;
{
  static char	host[132];	/* Unix domain is largest */
  char *	retval = host;

  memset(host, 0, sizeof(host));
  switch (sa0->sa_family) {
    case 255: {
	unsigned long	mask;
	unsigned int	index0 = 1 << 31;
	unsigned short	new_mask = 0;

	mask = ntohl(((const struct sockaddr_in *)sa0)->sin_addr.s_addr);
	while (mask & index0) {
	  new_mask++;
	  index0 >>= 1;
	}
	if (snprintf(host, sizeof(host), "/%hu", new_mask) == -1) {
	  return NULL;
	}
      }
      break;
    case AF_UNSPEC:
    case AF_INET: {
	struct sockaddr_in *	sa = (struct sockaddr_in *)sa0;

	if (inet_ntop(AF_INET, &sa->sin_addr, host, sizeof(host)) != NULL) {
	  size_t	len = strlen(host);

	  if (ntohs(sa->sin_port) != 0 && with_port == true) {
	    char	port[7];
	    int		n;

	    n = snprintf(port, sizeof(port), ".%d", ntohs(sa->sin_port));
	    if (n != -1) {
	      strlcat(host, port, sizeof(host) - len);
	    }
	  }
	}
      }
      break;
    case AF_INET6: {
	struct sockaddr_in6 *	sa = (struct sockaddr_in6 *)sa0;

	if (inet_ntop(AF_INET6, &sa->sin6_addr, host, sizeof(host)) != NULL) {
	  size_t	len = strlen(host);

	  if (ntohs(sa->sin6_port) != 0 && with_port == true) {
	    char	port[7];
	    int		n;

	    n = snprintf(port, sizeof(port), ".%d", ntohs(sa->sin6_port));
	    if (n != -1) {
	      strlcat(host, port, sizeof(host) - len);
	    }
	  }
	}
      }
      break;
    case AF_UNIX: {
	struct sockaddr_un *	sa = (struct sockaddr_un *)sa0;

	/*
	 * OK to have no pathname bound to the socket: happens on every
	 * connect() unless client calls bind() first.
	 */
	if (sa->sun_path[0] == 0) {
	  strlcpy(host, "(no pathname bound)", sizeof(host));
	}
	else {
	  snprintf(host, sizeof(host), "%s", sa->sun_path);
	}
      }
      break;
    case AF_LINK: {
	struct sockaddr_dl *	sa = (struct sockaddr_dl *)sa0;

	if (sa->sdl_type == IFT_ETHER && sa->sdl_alen > 0) {
	  char	s[NETFD_ETHER_ADDR_SIZE];

	  etherprint((uint8_t *)LLADDR(sa), s);
	  strlcpy(host, s, sizeof(host));
	}
	else if (sa->sdl_nlen > 0) {
	  /* interface name */
	  memset(host, 0, sizeof(host));
	  memcpy(host, sa->sdl_data, sa->sdl_nlen);
	}
	else {
	  snprintf(host, sizeof(host), "<Link#%d>", sa->sdl_index);
	}
      }
      break;
    default:
      break;
  }

  if (dst != NULL) {
    strlcpy(dst, host, size);
  }

  return retval;
}
#endif

static char *
etherprint(ea, s)
	const uint8_t	ea[6];
	char		s[NETFD_ETHER_ADDR_SIZE];
{
  snprintf(s, NETFD_ETHER_ADDR_SIZE,
	   "%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx",
	   ea[0], ea[1], ea[2], ea[3], ea[4], ea[5]);
  s[NETFD_ETHER_ADDR_SIZE-1] = '\0';

  return s;
}

static int
get_ifconf(sd, ifc, buf)
	int		sd;
	struct ifconf *	ifc;	/* returned */
	char **		buf;	/* returned */
{
  size_t	lastlen	= 0;
  size_t	len;
  
  len = 100 * sizeof(struct ifreq);	/* initial buffer size guess */
  while (1) {
    *buf = malloc(len);
    if (*buf == NULL) {
      return -1;
    }
    ifc->ifc_len = len;
    ifc->ifc_buf = *buf;
    if (ioctl(sd, SIOCGIFCONF, ifc) < 0) {
      if (errno != EINVAL || lastlen != 0) {
	return -1;
      }
    }
    else {
      if ((size_t)ifc->ifc_len == lastlen) {
	break;		/* success, len has not changed */
      }
      lastlen = ifc->ifc_len;
    }
    len += 10 * sizeof(struct ifreq);	/* increment */
    free(*buf);
  }

  return 0;
}

netfd_ifinfo_t *
netfd_get_ifinfo(void)
{
  netfd_ifinfo_t *	ifihead = NULL;
  int			sd;

  sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
  if (sd != -1) {
    struct ifconf	ifc;
    char *		buf;

    if (get_ifconf(sd, &ifc, &buf) == 0) {
      netfd_ifinfo_t **	ifi_next;
      char		lastname[IFNAMSIZ];
      sa_family_t	lastfamily = AF_UNSPEC;
      char *		ptr;

      ifi_next	  = &ifihead;
      lastname[0] = '\0';

      for (ptr = buf; ptr < buf + ifc.ifc_len; ) {
	struct ifreq *		ifr;
	struct ifreq		ifrcopy;
	int			len;
	netfd_ifinfo_t *	ifi;
	int			flags;
	int			myflags;
	char *			cptr;
	sa_family_t		family;

	ifr  = (struct ifreq *)ptr;
	len  = max(sizeof(struct sockaddr), ifr->ifr_addr.sa_len);
	ptr += sizeof(ifr->ifr_name) + len;	/* for next one in buffer */

	myflags = 0;
	if ( (cptr = strchr(ifr->ifr_name, ':')) != NULL) {
	  *cptr = 0;		/* replace colon will null */
	}
	family = ifr->ifr_addr.sa_family;
	if (strncmp(lastname, ifr->ifr_name, IFNAMSIZ) == 0 &&
	    lastfamily == family) {
	  if (family == AF_INET || family == AF_INET6) {
	    myflags = IFI_ALIAS;
	  }
	}
	if (family != AF_LINK) {
	  memcpy(lastname, ifr->ifr_name, IFNAMSIZ);
	}
	lastfamily = family;

	ifrcopy = *ifr;
	ioctl(sd, SIOCGIFFLAGS, &ifrcopy);
	flags = ifrcopy.ifr_flags;
	if ((flags & IFF_UP) == 0) {
	  continue;	/* ignore if interface not up */
	}

	ifi = calloc(1, sizeof(struct ifi_info));
	if (ifi == NULL) { continue; }
	*ifi_next = ifi;		/* prev points to this new one */
	ifi_next  = &ifi->ifi_next;	/* pointer to next one goes here */

	ifi->ifi_flags	 = flags;	/* IFF_xxx values */
	ifi->ifi_myflags = myflags;	/* IFI_xxx values */
	memcpy(ifi->ifi_name, ifr->ifr_name, IFI_NAME);
	ifi->ifi_name[IFI_NAME-1] = '\0';

	switch (family) {
       	  case AF_INET: {
	      struct sockaddr_in *	sa;

	      sa = (struct sockaddr_in *)&ifr->ifr_addr;
	      if (ifi->ifi_addr == NULL) {
		ifi->ifi_addr = calloc(1, sizeof(struct sockaddr_in));
		if (ifi->ifi_addr != NULL) {
		  memcpy(ifi->ifi_addr, sa, sizeof(struct sockaddr_in));
		}

		if (flags & IFF_BROADCAST) {
		  ioctl(sd, SIOCGIFBRDADDR, &ifrcopy);
		  sa = (struct sockaddr_in *)&ifrcopy.ifr_broadaddr;
		  ifi->ifi_brdaddr = calloc(1, sizeof(struct sockaddr_in));
		  if (ifi->ifi_brdaddr != NULL) {
		    memcpy(ifi->ifi_brdaddr, sa, sizeof(struct sockaddr_in));
		  }
		}

		if (flags & IFF_POINTOPOINT) {
		  ioctl(sd, SIOCGIFDSTADDR, &ifrcopy);
		  sa = (struct sockaddr_in *)&ifrcopy.ifr_dstaddr;
		  ifi->ifi_dstaddr = calloc(1, sizeof(struct sockaddr_in));
		  if (ifi->ifi_dstaddr != NULL) {
		    memcpy(ifi->ifi_dstaddr, sa, sizeof(struct sockaddr_in));
		  }
		}
	      }
	    }
    	    break;
	  case AF_INET6: {
	      struct sockaddr_in6 *	sa;

	      sa = (struct sockaddr_in6 *)&ifr->ifr_addr;
	      if (ifi->ifi_addr == NULL) {
		ifi->ifi_addr = calloc(1, sizeof(struct sockaddr_in6));
		if (ifi->ifi_addr != NULL) {
		  memcpy(ifi->ifi_addr, sa, sizeof(struct sockaddr_in6));
		}
	      }
	    }
	    break;
	  case AF_LINK: {
	      struct sockaddr_dl *	sa;

	      sa = (struct sockaddr_dl *)&ifr->ifr_addr;
	      if (sa->sdl_type == IFT_ETHER && sa->sdl_alen > 0) {
		ifi->ifi_hlen = sa->sdl_alen;
		memset(ifi->ifi_haddr, 0, sizeof(ifi->ifi_haddr));
		memcpy(ifi->ifi_haddr, (uint8_t *)LLADDR(sa), sa->sdl_alen);
	      }
	      if (ifi->ifi_addr == NULL) {
		ifi->ifi_addr = calloc(1, sizeof(struct sockaddr_dl));
		if (ifi->ifi_addr != NULL) {
		  memcpy(ifi->ifi_addr, sa, sizeof(struct sockaddr_dl));
		}
	      }
	    }
	    break;
	  default:
    	    break;
	}
      }
      free(buf);
    }
    close(sd);
  }

  return ifihead;	/* pointer to first structure in linked list */
}

void
netfd_free_ifinfo(netfd_ifinfo_t * ifihead)
{
  netfd_ifinfo_t *	ifi;
  netfd_ifinfo_t *	ifi_next;

  for (ifi = ifihead; ifi != NULL; ifi = ifi_next) {
    if (ifi->ifi_addr    != NULL) { free(ifi->ifi_addr); }
    if (ifi->ifi_brdaddr != NULL) { free(ifi->ifi_brdaddr); }
    if (ifi->ifi_dstaddr != NULL) { free(ifi->ifi_dstaddr); }
    ifi_next = ifi->ifi_next;	/* can't fetch ifi_next after free() */
    free(ifi);			/* the ifi_info{} itself */
  }
}

/*****************************************************************************/

int
netfd_dispatch(callback, writeshot, timeout)
	netfd_callback_t	callback;
	const bool		writeshot;
				/* true: enable EV_WRITE|EV_ONESHOT */
	const time_t		timeout;	/* milliseconds */
{
  register int		i;
  int			n;
  struct timespec	to;
  struct timespec *	tsp;

  if (dispatched) { return -1; }

  if (timeout == 0) {
    to.tv_sec	= 0L;
    to.tv_nsec	= 0L;
    tsp		= &to;
  }
  else if (timeout < 0) {
    tsp = NULL;
  }
  else {
    to.tv_sec	= NETFD_MSEC2SEC(timeout);
    to.tv_nsec	= NETFD_MSEC2NSEC(timeout);
    tsp		= &to;
  }

  dispatched = 1;
  do {
    n = kevent(nvp->kq.fd, nvp->kq.events, nvp->kq.nevents,
	       nvp->kq.events, nvp->maxfds, tsp);
    switch (n) {
      case -1: break;
      case 0:  errno = ETIMEDOUT; break;
      default:
	nvp->kq.nevents = 0;
	for (i = 0; i < n; i++) {
	  NetFD	nfd = (NetFD)nvp->kq.events[i].udata;

	  if (nfd->flags == NETFD_FL_LISTENER) {
	    struct sockaddr 	cliaddr;
	    socklen_t		clilen;
	    int			sd;

	    sd = accept(nfd->osfd, &cliaddr, &clilen);
	    if (sd != -1) {
	      NetFD	newfd;
#ifdef	SO_NOSIGPIPE
	      const int	on = 1;
	      /* Do not generate SIGPIPE */
	      (void)setsockopt(sd, SOL_SOCKET, SO_NOSIGPIPE, &on, sizeof(on));
#endif	/* SO_NOSIGPIPE */

	      newfd = netfd_fdopen(sd);
	      if (newfd != NULL) {
		newfd->saddr = (struct sockaddr *)malloc(clilen);
		if (newfd->saddr != NULL) {
		  memcpy(newfd->saddr, &cliaddr, clilen);
		  newfd->addrlen = clilen;
		}
		else {
		  netfd_close(newfd);
		  break;	/* exit switch */
		}
		EV_SET(&nvp->kq.events[nvp->kq.nevents++], newfd->osfd,
		       EVFILT_READ, EV_ADD, 0, 0, newfd);
		/*
		 * Generally, server writes data after it reads data.
		 * This writeshot flag enable server to write data immediately
		 * without reading any data.
		 */
		if (writeshot) {
		  EV_SET(&nvp->kq.events[nvp->kq.nevents++], newfd->osfd,
			 EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, newfd);
		}
	      }
	      else {
		my_close(sd);
	      }
	    }
	  }
	  else if (nvp->kq.events[i].flags == EV_EOF) {
	    netfd_close(nfd);
	  }
	  else {
#if	DEBUG
	    nfd->count++;
#endif	/* DEBUG */
	    (*callback)(nfd);
	  }
	}
	break;
    }
  } while (n > 0);
  dispatched = 0;

  return n;	/* -1: error, 0: timeout */
}

void
netfd_event_set(nfd, filter, flags)
	NetFD	nfd;
	int		filter;
	int		flags;
{
  EV_SET(&nvp->kq.events[nvp->kq.nevents++], nfd->osfd, filter, flags,
	 0, 0, nfd);
}

/******************************************************************************
 *
 *	ICMP Echo and Reply functions
 *
 *****************************************************************************/
static uint16_t	in_cksum(uint16_t *, size_t);
static void	tv_sub(struct timeval *, struct timeval *);
static int	proc_icmp(uint8_t * buf, ssize_t len, uint16_t check_id,
			  struct sockaddr * from, struct timeval * tvrecv);

#define	PHDR_LEN	((int)sizeof(struct timeval))
#define	DEFDATALEN	(64 - PHDR_LEN)	/* default data length */
#define	MAXIPLEN	(sizeof(struct ip) + MAX_IPOPTLEN)
#define	MINICMPLEN	ICMP_MINLEN
#define	MAXPAYLOAD	(IP_MAXPACKET - MAXIPLEN - MINICMPLEN)

NetFD
netfd_icmp_open(void)
{
  int	sd;

  sd = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
  if (sd != -1) {
    static const int	size = IP_MAXPACKET + 128;
    NetFD		nfd;

    /* OK even if setsockopt fails */
    (void)setsockopt(sd, SOL_SOCKET, SO_RCVBUF, &size, sizeof(size));
    if (getuid() == 0) {
      (void)setsockopt(sd, SOL_SOCKET, SO_SNDBUF, &size, sizeof(size));
    }
    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

ssize_t
netfd_icmp_send(nfd, to, tolen, send_id, seqno, timeout)
	NetFD			nfd;
	const struct sockaddr *	to;
	socklen_t		tolen;
	uint16_t		send_id;	/* icmp_id */
	uint32_t		seqno;		/* icmp_seq */
	const time_t		timeout;	/* milliseconds */
{
  struct icmp *	icmp;
  uint8_t	sendbuf[MINICMPLEN + MAXPAYLOAD];
  size_t	len;

  /* make ICMP packet */
  icmp = (struct icmp *)sendbuf;
  icmp->icmp_type	= ICMP_ECHO;
  icmp->icmp_code	= 0;
  icmp->icmp_cksum	= 0;
  icmp->icmp_id		= send_id;
  icmp->icmp_seq	= seqno;
  gettimeofday((struct timeval *)icmp->icmp_data, NULL);

  len = MINICMPLEN + DEFDATALEN;	/* checksum ICMP header and data */
  icmp->icmp_cksum = in_cksum((uint16_t *)icmp, len);

  return netfd_sendto(nfd, sendbuf, (size_t)len, 0, to, tolen, timeout);
}

static uint16_t
in_cksum(addr, len)
	uint16_t *	addr;
	size_t		len;
{
  int		nleft	= len;
  int		sum	= 0;
  uint16_t *	w	= addr;
  uint16_t	answer	= 0;

  /*
   * Our algorithm is simple, using a 32 bit accumulator (sum),
   * we add sequential 16 bit words to it, and at the end,
   * fold back all the carry bits from the top 16 bits into the lower 16 bits.
   */
  while (nleft > 1) {
    sum	  += *w++;
    nleft -= 2;
  }

  /* 4mop up an odd byte, if neccessary */
  if (nleft == 1) {
    *(uint8_t *)(&answer) = *(uint8_t *)w;
    sum += answer;
  }

  sum	 = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
  sum	+= (sum >> 16);				/* add carry */
  answer = ~sum;				/* truncate to 16 bits */

  return answer;
}

int
netfd_icmp_recv(nfd, from, fromlen, check_id, timeout, callback)
	NetFD			nfd;
	struct sockaddr *	from;
	socklen_t *		fromlen;
	uint16_t		check_id;	/* icmp id */
	const time_t		timeout;
	netfd_icmp_callback_t	callback;
{
  uint8_t	recvbuf[MINICMPLEN + MAXPAYLOAD];
  ssize_t	n;

  n = netfd_recvfrom(nfd, recvbuf, sizeof(recvbuf), 0, from, fromlen, timeout);
  if (n > 0) {
    struct timeval	tval;

    gettimeofday(&tval, (struct timezone *)NULL);
    if (callback == NULL) {
      callback = proc_icmp;
    }
    n = (*callback)(recvbuf, n, check_id, from, &tval);
  }
#if	DEBUG0
  fprintf(stderr, "DEBUG[netfd_icmp_recv] n=%d errno=%d\n", n, errno);
#endif	/* DEBUG */

  return n;
}

static int
proc_icmp(buf, len, check_id, from, tvrecv)
	uint8_t *		buf;
	ssize_t			len;
	uint16_t		check_id;	/* checked icmp id */
	struct sockaddr *	from;
	struct timeval *	tvrecv;
{
  const struct ip *	ip;
  int			hlen;
  const struct icmp *	icmp;
  int			icmplen;

  ip	= (const struct ip *)buf;	/* start of IP header */
  hlen	= ip->ip_hl << 2;		/* length of IP header */

  icmp	= (const struct icmp *)(buf + hlen);	/* start of ICMP header */
  icmplen = len - hlen;
  if (icmplen < 8) {
    fprintf(stderr, "icmplen (%d) < 8\n", icmplen);
    return -1;
  }

  if (icmp->icmp_type == ICMP_ECHOREPLY) {
    struct timeval *	tvsend;
    double		rtt;

    if (icmp->icmp_id != check_id) {
      return -1;	/* not a response to our ECHO_REQUEST */
    }
    if (icmplen < 16) {
      fprintf(stderr, "icmplen (%d) < 16\n", icmplen);
      return -1;
    }

    tvsend = (struct timeval *)(icmp->icmp_data);
    tv_sub(tvrecv, tvsend);
    rtt	= tvrecv->tv_sec * 1000.0 + tvrecv->tv_usec / 1000.0;

    fprintf(stdout, "%d bytes from %s: seq=%u ttl=%d rtt=%.3f ms\n",
	    icmplen, netfd_sock_ntop(from, false, NULL, 0),
	    icmp->icmp_seq, ip->ip_ttl, rtt);
    return 0;
  }

  return -1;
}

static void
tv_sub(out, in)
	register struct timeval *	out;
	register struct timeval *	in;
{
  out->tv_usec -= in->tv_usec;
  if (out->tv_usec < 0) {
    --out->tv_sec;
    out->tv_usec += 1000000;
  }
  out->tv_sec -= in->tv_sec;
}

/******************************************************************************
 *
 *	Raw IP and BPF manipulation functions
 *
 *****************************************************************************/
/*
 * RFC 793
 * +--------+--------+--------+--------+
 * |           Source Address          |
 * +--------+--------+--------+--------+
 * |         Destination Address       |
 * +--------+--------+--------+--------+
 * |  zero  |  PTCL  | TCP/UDP Length  |
 * +--------+--------+--------+--------+
 */
struct pseudo_hdr {	/* TCP pseudo header for checksum */
  in_addr_t	saddr;
  in_addr_t	daddr;
  uint8_t	zero;
  uint8_t	protocol;
  uint16_t	length;
};
#define	SIZE_ETHERHDR	(sizeof(struct ether_header))
#define	SIZE_IPHDR	(sizeof(struct ip))
#define	SIZE_TCPHDR	(sizeof(struct tcphdr))
#define	SIZE_UDPHDR	(sizeof(struct udphdr))
#define	SIZE_ICMPHDR	(sizeof(struct icmp))
#define	SIZE_BPFHDR	(sizeof(struct bpf_hdr))
#define	SIZE_PSEUDOHDR	(sizeof(struct pseudo_hdr))

NetFD
netfd_open_raw_ip(has_hincl)
	bool	has_hincl;
{
  int	sd;

  sd = socket(PF_INET, SOCK_RAW, IPPROTO_RAW);
  if (sd != -1) {
    static const int	size  = IP_MAXPACKET;
    static const int	hincl = 1;
    int			rv    = 0;

    (void)setsockopt(sd, SOL_SOCKET, SO_RCVBUF, &size, sizeof(size));
    (void)setsockopt(sd, SOL_SOCKET, SO_SNDBUF, &size, sizeof(size));
    if (has_hincl == true) {
      rv = setsockopt(sd, IPPROTO_IP, IP_HDRINCL, &hincl, sizeof(hincl));
    }
    if (rv == 0) {
      NetFD	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

NetFD
netfd_open_bpf(netif, flags, capsize, promiscuous, to_ms, pktfilter)
	const char *		netif;		/* network interface */
	uint16_t		flags;		/* bpf flags */
	size_t			capsize;	/* buffer size to capture */
	bool			promiscuous;	/* promiscus: true|false */
	const time_t		to_ms;		/* timeout (msec) */
	struct bpf_program *	pktfilter;	/* packet filter */
{
  if (netif != NULL) {
    int	fd = -1;
    int	n  = 0;

    do {
      char	s[sizeof("/dev/bpfXXX")];

      (void)snprintf(s, sizeof(s), "/dev/bpf%d", n++);
      fd = open(s, O_RDWR);
    } while (fd == -1 && (errno == EBUSY || errno == EPERM));

    if (fd != -1) {
      struct ifreq	ifreq;

      /* set buffer size to capture */
      if (capsize > 0) {
	size_t	bufsize;

	for (bufsize = capsize; bufsize != 0; bufsize >>= 1) {
	  if (ioctl(fd, BIOCSBLEN, &bufsize) == 0) { break; }
	}
	if (bufsize == 0) { goto error; }
      }
      /* set network device */
      memset(&ifreq, 0, sizeof(ifreq));
      strlcpy(ifreq.ifr_name, netif, sizeof(ifreq.ifr_name));
      if ( ioctl(fd, BIOCSETIF, &ifreq) == -1) { goto error; }
      /* set packet filter */
      if (pktfilter != NULL) {
	if (ioctl(fd, BIOCSETF, pktfilter) == -1) { goto error; }
      }
      /* set timeout */
      if (to_ms > 0) {
	struct timeval	timeout;

	timeout.tv_sec  = NETFD_MSEC2SEC(to_ms);
	timeout.tv_usec = NETFD_MSEC2USEC(to_ms);
	if (ioctl(fd, BIOCSRTIMEOUT, &timeout) == -1) { goto error; }
      }
      /* set promiscus mode */
      if (promiscuous) {
       	if (ioctl(fd, BIOCPROMISC, NULL) == -1) { goto error; }
      }
      /* no block in kernel buffer */
      if (flags & NETFD_BPF_IMMEDIATE) {
	static const unsigned int	on = 1;
	if (ioctl(fd, BIOCIMMEDIATE, &on) == -1) { goto error; }
      }
      /* build ether frame by yourself */
      if (flags & NETFD_BPF_HDRCMPLT) {
	static const unsigned int	on = 1;
	if (ioctl(fd, BIOCSHDRCMPLT, &on) == -1) { goto error; }
      }
      /* see only received packets */
      if (flags & NETFD_BPF_NOSEESENT) {
	static const unsigned int	off = 0;
	if (ioctl(fd, BIOCSSEESENT, &off) == -1) { goto error; }
      }
      if (ioctl(fd, BIOCFLUSH, 0) == 0) {
	NetFD	nfd;

	nfd = netfd_fdopen(fd);
	if (nfd != NULL) {
	  return nfd;
	}
      }
error:
      my_close(fd);
    }
  }

  return NULL;
}

/*****************************************************************************/

ssize_t
netfd_build_tcp_packet(src_addr, src_port, dst_addr, dst_port,
		       flags, seqno, ackno, wsize, ip_id, packet)
	const char *	src_addr;	/* source address */
	uint16_t	src_port;	/* source port */
	const char *	dst_addr;	/* destination address */
	uint16_t	dst_port;	/* destination port */
	uint8_t		flags;		/* tcp flags */
	uint32_t	seqno;		/* tcp sequence number */
	uint32_t	ackno;		/* tcp acknowledge number */
	uint16_t	wsize;		/* tcp window size */
	uint8_t		ip_id;		/* identification */
	uint8_t **	packet;
{
  if (packet != NULL) {
    uint8_t *	tcppacket;
    size_t	packetsize;
    size_t	tcpoptsize = 0;

    if ((flags & TH_SYN) == TH_SYN) {
      tcpoptsize = 12;	/* for timestamp option (see RFC1323) */
    }

    packetsize = SIZE_TCPHDR + tcpoptsize;
    tcppacket  = (uint8_t *)malloc(SIZE_PSEUDOHDR + packetsize);
    if (tcppacket != NULL) {
      const in_addr_t		saddr = inet_addr(src_addr);
      const in_addr_t		daddr = inet_addr(dst_addr);
      struct pseudo_hdr *	pseudo;
      struct tcphdr *		tcp;
      ssize_t			retval;

      pseudo = (struct pseudo_hdr *)tcppacket;
      tcp    = (struct tcphdr *)(tcppacket + SIZE_PSEUDOHDR);

      memset(tcppacket, 0, SIZE_PSEUDOHDR + packetsize);
      /* build pseudo header */
      memcpy(&pseudo->saddr, &saddr, sizeof(in_addr_t));
      memcpy(&pseudo->daddr, &daddr, sizeof(in_addr_t));
      pseudo->protocol = IPPROTO_TCP;
      pseudo->length   = htons(packetsize);

      /* build header */
      tcp->th_sport = htons(src_port);
      tcp->th_dport = htons(dst_port);
      tcp->th_seq   = htonl(seqno);
      tcp->th_ack   = htonl(ackno);
      tcp->th_x2    = 0;
      tcp->th_off   = (SIZE_TCPHDR >> 2) + (tcpoptsize >> 2);
      tcp->th_flags = flags;
      tcp->th_win   = htons(wsize);
      tcp->th_urp   = 0;

      /* tcp timestamp option */
      if (tcpoptsize) {
	uint8_t *	tstamp;
	time_t		now;

	tstamp = (uint8_t *)(tcppacket + SIZE_PSEUDOHDR + SIZE_TCPHDR);
	tstamp[0] = 8;
	tstamp[1] = 10;	/* 10bytes = kind + length + T1 + T2 */
	time(&now);
	memcpy(tstamp + 2, &now, 4);
	memset(tstamp + 6, 0, 4);	/* zero clear */
	tstamp[10] = 1;	/* NOP */
	tstamp[11] = 0;	/* EOL */
      }

      /* calculate checksum */
      tcp->th_sum  = in_cksum((uint16_t *)tcppacket,
			      SIZE_PSEUDOHDR + packetsize);

      /* build ip packet */
      retval = netfd_build_ip_packet(IPPROTO_TCP, src_addr, dst_addr,
				     tcppacket + SIZE_PSEUDOHDR, packetsize,
				     IPDEFTTL, 0, IP_DF, ip_id, packet);
      if (retval > 0) {
	free(tcppacket);
      }

      return retval;
    }
  }

  return -1;
}

ssize_t
netfd_build_udp_packet(src_addr, src_port, dst_addr, dst_port, ip_id, packet)
	const char *	src_addr;	/* source address */
	uint16_t	src_port;	/* source port */
	const char *	dst_addr;	/* destination address */
	uint16_t	dst_port;	/* destination port */
	uint8_t		ip_id;		/* identification */
	uint8_t **	packet;
{
  if (packet != NULL) {
    uint8_t *	udppacket;
    size_t	packetsize;

    packetsize = SIZE_UDPHDR;
    udppacket  = (uint8_t *)malloc(SIZE_PSEUDOHDR + packetsize);
    if (udppacket != NULL) {
      const in_addr_t		saddr = inet_addr(src_addr);
      const in_addr_t		daddr = inet_addr(dst_addr);
      struct pseudo_hdr *	pseudo;
      struct udphdr *		udp;
      ssize_t			retval;

      pseudo = (struct pseudo_hdr *)udppacket;
      udp    = (struct udphdr *)(udppacket + SIZE_PSEUDOHDR);

      memset(udppacket, 0, SIZE_PSEUDOHDR + packetsize);
      /* build pseudo header */
      memcpy(&pseudo->saddr, &saddr, sizeof(in_addr_t));
      memcpy(&pseudo->daddr, &daddr, sizeof(in_addr_t));
      pseudo->protocol = IPPROTO_UDP;
      pseudo->length   = htons(packetsize);

      /* build header */
      udp->uh_sport = htons(src_port);
      udp->uh_dport = htons(dst_port);
      udp->uh_ulen  = htons(packetsize);

      /* calculate checksum */
      udp->uh_sum   = in_cksum((uint16_t *)udppacket,
			       SIZE_PSEUDOHDR + packetsize);

      /* build ip packet */
      retval = netfd_build_ip_packet(IPPROTO_UDP, src_addr, dst_addr,
				     udppacket + SIZE_PSEUDOHDR, packetsize,
				     IPDEFTTL, 0, IP_DF, ip_id, packet);
      if (retval > 0) {
	free(udppacket);
      }

      return retval;
    }
  }

  return -1;
}

ssize_t
netfd_build_ip_packet(protocol, src_addr, dst_addr, data, datasize,
		      ip_ttl, ip_off, ip_tos, ip_id, packet)
	uint8_t		protocol;	/* protocol number */
	const char *	src_addr;	/* source address */
	const char *	dst_addr;	/* destination address */
	uint8_t *	data;		/* paylaod */
	size_t		datasize;	/* payload size */
	uint8_t		ip_ttl;		/* time to live */
	uint16_t	ip_off;		/* fragment offset field */
	uint8_t		ip_tos;		/* type of service */
	uint8_t		ip_id;		/* identification */
	uint8_t **	packet;		/* returned */
{
  if (packet != NULL) {
    size_t	packetsize;

    packetsize = SIZE_IPHDR + datasize;
    *packet    = (uint8_t *)malloc(packetsize);
    if (*packet != NULL) {
      const in_addr_t	saddr = inet_addr(src_addr);
      const in_addr_t	daddr = inet_addr(dst_addr);
      struct ip *	ip;

      memset(*packet, 0, packetsize);
      ip = (struct ip *)(*packet);

      /* build header */
      ip->ip_v	 = IPVERSION;
      ip->ip_hl	 = SIZE_IPHDR >> 2;
      ip->ip_tos = ip_tos;
      ip->ip_len = htons(packetsize);
      ip->ip_id	 = htons(ip_id);
      ip->ip_off = htons(ip_off);
      ip->ip_ttl = ip_ttl;
      ip->ip_p	 = protocol;
      ip->ip_sum = 0;		/* zero clear to recompute */
      ip->ip_src = *(const struct in_addr *)&saddr;
      ip->ip_dst = *(const struct in_addr *)&daddr;

      /* compute checksum */
      ip->ip_sum = in_cksum((uint16_t *)*packet, SIZE_IPHDR);

      /* copy payload */
      memcpy(*packet + SIZE_IPHDR, data, datasize);

      return packetsize;
    }
  }

  return -1;
}

ssize_t
netfd_build_ether_frame(type, src_host, dst_host, data, datasize, frame)
	uint16_t	type;		/* ether type */
	const char *	src_host;	/* source host address */
	const char *	dst_host;	/* destination host address */
	uint8_t *	data;		/* paylaod */
	size_t		datasize;	/* payload size */
	uint8_t **	frame;		/* returned */
{
  if (frame != NULL) {
    size_t	framesize;

    framesize = SIZE_ETHERHDR + datasize;
    *frame    = (uint8_t *)malloc(framesize);
    if (*frame != NULL) {
      struct ether_header *	ether;
      struct ether_addr		shost;
      struct ether_addr		dhost;

      memset(*frame, 0, framesize);
      ether = (struct ether_header *)(*frame);

      /* build header */
      shost = *(struct ether_addr *)ether_aton(src_host);
      dhost = *(struct ether_addr *)ether_aton(dst_host);
      ether->ether_type = htons(type);
      memcpy(ether->ether_shost, shost.octet, ETHER_ADDR_LEN);
      memcpy(ether->ether_dhost, dhost.octet, ETHER_ADDR_LEN);

      /* copy payload */
      memcpy(*frame + SIZE_ETHERHDR, data, datasize);

      return framesize;
    }
  }

  return -1;
}
